#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define DIM 20

// void BoundaryConditions(int msize, double *u)
void BoundaryConditions(int msize, double u[][DIM])
{
// PDE: Laplacian u = 0;      0<=x<=1;  0<=y<=1
// B.C.: u(x,0)=sin(pi*x); u(x,1)=sin(pi*x)*exp(-pi); u(0,y)=u(1,y)=0
// SOLUTION: u(x,y)=sin(pi*x)*exp(-pi*y)
  int i;
  double exp_pi = exp(M_PI);

  
  // u(x,0)=sin(pi*x)
  for (i = 0; i < msize; i++)
    // u[msize*i+0] = sin(2*M_PI*i/(double) msize);
    u[i][0] = sin(M_PI*i/(double) msize);

  // u(x,1)=sin(pi*x)*exp(-pi)
  for (i = 0; i < msize; i++)
    // u[msize*i + msize-1] = u[msize*i+0]*exp_pi;
    u[i][msize-1] = u[i][0]*exp_pi;

  // u(0,y)=u(1,y)=0
  for (i = 0; i < msize; i++) {
    u[0][i] = 0.0;
    // u[msize*(msize-1)+i] = 0.0;
    u[msize-1][i] = 0.0;
  }
}


int main () {

    int msize = DIM;
    int maxIter;
    double u1[DIM][DIM];  // solution array
    double u2[DIM][DIM];
    // double *oldSolution; // switching pointers to u1 and u2
    // double *newSolution;
    double (*newSolution)[DIM];
    double (*oldSolution)[DIM];
    double *x, *c, *n, *e, *w, *s; // pointers to single elements
    double maxvariation, variation=1.0e+4, tollerance=1.0e-4;

    FILE *fp;
    
    // loop counters
    int iter=0, i, j;
    
    i = maxIter = 0;
    while (maxIter <= 0 || i != 1 ) {
        printf("Enter max iterations, (maxIter): ");
        i = scanf(" %d", &maxIter);
    }
    fprintf(stderr, "Running Laplace Solver for maxIter: %d\n", maxIter);
    
    oldSolution = u1;
    newSolution = u2;
    // zeroing all elements
    for (i = 0; i < msize; i++)
     for (j = 0; j < msize; j++)
      oldSolution[i][j] = 0.0;

    // set up boundary values
    BoundaryConditions(msize, oldSolution);

    // iterate until error is below threshold
    do { 

      iter = iter + 1;

      if (iter > maxIter) {
        printf("Program terminated (Iteration exceeds %d)\n", maxIter);
        break;
      }

      maxvariation = 0.0;
      for (i = 1; i < msize-1; i++) {
        for (j = 1; j < msize-1; j++) {
          double var;

          // selecting current element and neighbours
          // x = &newSolution[msize*i+j];
          // c = &oldSolution[msize*i+j];
          // n = c+msize; s = c-msize;
          // e = c+1; w = c-1;
          x = &newSolution[i][j];
          c = &oldSolution[i][j];
          n = c+msize; s = c-msize;
          e = c+1; w = c-1;

          // new solution
          *x = 0.25 * ( *n + *e + *w + *s );

          // compute variation
          var = fabs(*c - *x);
          if (maxvariation < var) maxvariation = var;
        }
      }

      // swap 
      if ( iter%2 == 0 ) {
        oldSolution = u1;
        newSolution = u2;
      } else {
        oldSolution = u2;
        newSolution = u1;
      }

      if(iter%1000==0) 
            printf("%d %12.4f\n", iter, variation);

      if (variation > maxvariation) variation = maxvariation;

    } while (variation > tollerance);

    printf("Stopped at iteration %d\n", iter);
    printf("Maximum error is %lf\n\n", variation);

    // write results
    fp = fopen("results.dat", "w");
    if ( fp == NULL ) {
        printf("\nError in opening results.dat file\n");
        exit(-1);
    }

    for (i = 0; i < msize; i++) {
      for (j = 0;j < msize; j++) {
         // fprintf(fp, "%6d %6d %18.9f\n", i, j, oldSolution[msize*i+j]);
         fprintf(fp, "%6d %6d %18.9f\n", i, j, oldSolution[i][j]);
      }
    }
    fclose(fp);
    
    return 0;
}
